"""Store celery task id

Revision ID: a855c39e2a0f
Revises: 82172e6a3154
Create Date: 2020-03-11 08:25:27.973866

"""

# revision identifiers, used by Alembic.
revision = "a855c39e2a0f"
down_revision = "82172e6a3154"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("celery_task_id", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "celery_task_id")
    # ### end Alembic commands ###
