"""Add index to Compose.state

Revision ID: 0571a5ca58a0
Revises: 0d4d8e1cfe29
Create Date: 2017-09-21 10:29:48.025577

"""

# revision identifiers, used by Alembic.
revision = "0571a5ca58a0"
down_revision = "0d4d8e1cfe29"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f("ix_composes_state"), "composes", ["state"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_composes_state"), table_name="composes")
    # ### end Alembic commands ###
