"""Add index to Compose.time_to_expire

Revision ID: 3b92820da295
Revises: 0571a5ca58a0
Create Date: 2017-09-21 11:03:04.006824

"""

# revision identifiers, used by Alembic.
revision = "3b92820da295"
down_revision = "0571a5ca58a0"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        op.f("ix_composes_time_to_expire"), "composes", ["time_to_expire"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_composes_time_to_expire"), table_name="composes")
    # ### end Alembic commands ###
