"""Add Compose.removed_by

Revision ID: f4bc999818d7
Revises: 11b350234051
Create Date: 2018-02-23 17:17:46.664813

"""

# revision identifiers, used by Alembic.
revision = "f4bc999818d7"
down_revision = "11b350234051"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("composes", sa.Column("removed_by", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("composes", "removed_by")
    # ### end Alembic commands ###
