<!DOCTYPE html>
<html lang="en">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
    <title>On Demand Compose Service</title>
  </head>
  <body>
    <h1>On Demand Compose Service</h1>
    <p>
    The ODCS (On Demand Compose Service) is a service that generates temporary
    composes using REST API calls. It is an open source project available at
    <a href="https://pagure.io/odcs">https://pagure.io/odcs</a>.
    </p>
    <h3><a href="/composes">Composes List</a></h3>
  </body>
</html>
