"""Create user model

Revision ID: 0d4d8e1cfe29
Revises: 566733ac3811
Create Date: 2017-07-06 22:37:38.576275

"""

# revision identifiers, used by Alembic.
revision = "0d4d8e1cfe29"
down_revision = "566733ac3811"

from alembic import op
import sqlalchemy as sa


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "users",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(length=200), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("users")
    # ### end Alembic commands ###
